
local MAJOR = "StarLibMarquee-1.0" 
local MINOR = 1
assert(LibStub, MAJOR.." requires LibStub") 
local LibMarquee = LibStub:NewLibrary(MAJOR, MINOR)
assert(LibMarquee, "LibStub.NewLibrary failed")
local LibProperty = LibStub("StarLibProperty-1.0")
assert(LibProperty, MAJOR .. " requires StarLibProperty-1.0")
local LibTimer = LibStub("StarLibTimer-1.0")
assert(LibTimer, MAJOR .. " requires StarLibTimer-1.0")
local error = LibStub("StarLibError-1.0")
assert(error, MAJOR .. " requires StarLiberror-1.0")
error = error:New(MAJOR)

local PINGPONGWAIT = 2
local ALIGN_LEFT, ALIGN_CENTER, ALIGN_RIGHT, ALIGN_MARQUEE, ALIGN_AUTOMATIC, ALIGN_PINGPONG = 1, 2, 3, 4, 5, 6
local SCROLL_RIGHT, SCROLL_LEFT = 1, 2
local aligns = {L = 1, C = 2, R = 3, M = 4, A = 5, P = 6}


if not LibMarquee.__index then
	LibMarquee.__index = LibMarquee
	LibMarquee.pool = setmetatable({}, {__mode = "k"})
end

local function resizeText(str, cols)
	if strlen(str) < cols then
		for i = strlen(str) + 1, cols do
			str = str .. " "
		end
	end
	return str:sub(1, cols)
end

local function replace(text1, pos1, text2, pos2)
	local char = text2:sub(pos2, pos2)
	
	local left = text1:sub(1, pos1 - 1)
	local right = text1:sub(pos1 + 1)
	
	return left .. char .. right
end

LibMarquee.defaults = {
	precision = 0xbabe,
	align = ALIGN_LEFT,
	update = 1000,
	speed = 500,
	direction = SCROLL_RIGHT,
	cols = 20
}

function LibMarquee:New(fontString, env, context) 

	assert(fontString)
	assert(env)
	assert(context)
	assert(context.name)
	assert(context.string)
	
	local obj = next(self.pool)

	if obj then
		self.pool[obj] = nil
	else
		obj = {}
	end
	
	setmetatable(obj, self)

	obj.value = LibProperty:New(env, context.name .. " string", context.string, "") -- text of marquee
	obj.prefix = LibProperty:New(env, context.name .. " prefix", context.prefix, "") -- label on the left side
	obj.postfix = LibProperty:New(env, context.name .. " postfix", context.string, "") -- label on right side
	obj.precision = context.precision or self.defaults.precision -- number of digits after the decimal point
	obj.align = aligns[context.align] or self.defaults.align -- alignment: left, center, right, marquee, automatic
	obj.update = context.update or self.defaults.update -- update interval
	obj.scroll = 0 -- marquee starting point
	obj.speed = context.speed or self.defaults.speed -- marquee scrolling speed
	obj.direction = context.direction or self.defaults.direction -- pingpong direction, 0 = right, 1 = left
	obj.cols = context.length or self.defaults.cols -- number of colums in marquee
	obj.offset = 0 -- increment by pixel
	obj.string = "" -- formatted value
	obj.fontString = fontString
	obj.timer = LibTimer:New(obj.update / 100, true, self.TextUpdate, obj)
	obj.textTimer = LibTimer:New(obj.speed / 100, true, self.TextScroll, obj)
	
	error:Print("test")
	return obj	
end

function LibMarquee:Del(ev)
	LibMarquee.pool[ev] = true
end

function LibMarquee:Start()
	error:Print("LibMarquee:Start")
	self.timer:Start()
	self.textTimer:Start()
end

function LibMarquee:Stop()
	StarTip:Print("LibMarquee:Stop")
	self.timer:Stop()
	self.textTimer:Stop()
end

function LibMarquee:Draw(text)
	self.fontString:SetText(text)
end

function LibMarquee:TextScroll()
	if not self.prefix or not self.postfix then return end
	local pre = self.prefix:P2S()
	local post = self.postfix:P2S()

	local str = self.string

	local num, len, width, pad
	local srcPtr, dstPtr = 0, 0
	local src = ""
	local dst = ""
	

    num = 0;
    len = strlen(str)
    width = self.cols - strlen(pre) - strlen(post);
    if width < 0 then
        width = 0
	end

    if self.direction == SCROLL_RIGHT and (self.align == ALIGN_MARQUEE or self.align == ALIGN_PINGPONG or self.align == ALIGN_AUTOMATIC) then
		self.offset = self.offset + 1
    elseif (self.align == ALIGN_MARQUEE or self.align == ALIGN_PINGPONG or self.align == ALIGN_AUTOMATIC) then
		self.offset = self.offset - 1
	end

    if(abs(self.offset) <= self.cols and (self.align == ALIGN_MARQUEE or self.align == ALIGN_PINGPONG or (self.align == ALIGN_AUTOMATIC and len > width))) then
		self:Draw(self.buffer or self.value:P2S())
        return
    else
        self.offset = 0;
    end

	if self.align == ALIGN_LEFT then
		pad = 0
	elseif self.align == ALIGN_CENTER then
		pad = (width - len) / 2
		if pad < 0 then
			pad = 0
		end
	elseif self.align == ALIGN_RIGHT then
		pad = width - len
		if pad < 0 then
			pad = 0
		end
	elseif self.align == ALIGN_AUTOMATIC then
		if len <= width then
			pad = 0
		end
	elseif self.align == ALIGN_MARQUEE then
		pad = width - self.scroll
		self.scroll = self.scroll + 1
		if self.scroll >= width + len then
			self.scroll = self.cols
		end
	elseif self.align == ALIGN_PINGPONG then
		if len <= width then
			pad = (width - len) / 2
		else
			if self.direction == SCROLL_RIGHT then
				self.scroll = self.scroll + 1
			else
				self.scroll = self.scroll - 1
			end
			
			pad = 0 - scroll_;
			
			if pad < 0 - (len - width) then
				if self.delay < 1 then
					self.direction = SCROLL_LEFT
					self.delay = PINGPONGWAIT
					self.scroll = self.scroll - PINGPONGWAIT
				end
				pad = 0 - (len - width)
			elseif pad > 0 then
				if self.delay < 1 then
					self.direction = SCROLL_RIGHT
					self.delay = PINGPONGWAIT
					self.scroll = self.scroll + PINGPOINGWAIT
				end
				pad = 0
			else
				pad = 0
			end
		end
	else
		pad = 0
		error:Print("No alignment specified")
	end

	dst = resizeText(dst, self.cols)
	
    dstPtr = 0;
    num = 0;

    -- /* process prefix */
    src = pre;
    while (num < self.cols) do
        if (srcPtr == strlen(src)) then
            break
		end
		dst = replace(dst, dstPtr, src, srcPtr)
        --dst[dstPtr] = src[srcPtr];
		dstPtr = dstPtr + 1
		srcPtr = srcPtr + 1
        num = num + 1
    end

    src = str;
    srcPtr = 0;
	
	local offset = pad
	
    if(offset < 0) then
        offset = 0;
	end
	
    -- /* wrap around on the beginning */
    while (pad > 0 and num < self.cols) do
        if(self.align == ALIGN_MARQUEE) then
			dst = replace(dst, dstPtr, src, (strlen(src) - offset) + srcPtr)
            --dst[dstPtr] = src[(strlen(src) - offset) + srcPtr];
			dstPtr = dstPtr + 1
			srcPtr = srcPtr + 1
        else
            dst[dstPtr] = ' ';
		end
		dstPtr = dstPtr + 1
        num = num + 1
        pad = pad - 1
    end

    --/* skip src chars (marquee) */
	local tmp = src
    while (pad < 0 and tmp ~= "") do
        src = tmp:sub(1); 
        tmp = src;
        pad = pad + 1
    end


    --/* copy content */
    while (num < self.cols) do
        if (srcPtr >= strlen(src)) then
            break;
		end
		dst = replace(dst, dstPtr, src, srcPtr)
        --dst[dstPtr] = src[srcPtr];
		dstPtr = dstPtr + 1
		srcPtr = srcPtr + 1
        num = num + 1
    end
	
    -- wrap around on end 
    src = post;
    len = strlen(src)
    srcPtr = 0;
    if(num < self.cols - len and self.align == ALIGN_MARQUEE) then
		dst = replace(dst, dstPtr, "*", 1)
        --dst[dstPtr] = '*';
		dstPtr = dstPtr + 1
        num = num + 1
    end
    while (num < self.cols - len) do
        if(self.align == ALIGN_MARQUEE) then
			dst = replace(dst, dstPtr, src, srcPtr)
            --dst[dstPtr] = str[srcPtr];
			dstPtr = dstPtr + 1
			srcPtr = srcPtr + 1
        else
			dst = replace(dst, dstPtr, ' ', 1)
            --dst[dstPtr] = ' ';
			dstPtr = dstPtr + 1
		end
        num = num + 1;
    end

    srcPtr = 0;

    --/* process postfix */
    while (num < self.cols) do
        if (srcPtr >= strlen(src)) then
            break;
		end
		dst = replace(dst, dstPtr, src, srcPtr)
        --dst[dstPtr] = src[srcPtr];
		dstPtr = dstPtr + 1
		srcPtr = srcPtr + 1
        num = num + 1
    end
	
	self.buffer = dst
	
	self:Draw(self.buffer)
end


function LibMarquee:TextUpdate()
	local str
	local update = 0
	
	if not self.prefix then return error:Print("LibMarquee needs a postfix")end
	if not self.postfix then error:Print("LibMarquee needs a prefix") end
	
	update = update + self.prefix:Eval()
	update = update + self.postfix:Eval()
	
    -- /* str or number? */
    if (self.precision == 0xBABE) then
        str = self.value:P2S();
    else
        local number = self.value:P2N();
        local width = self.cols - strlen(self.prefix:P2S()) - strlen(self.postfix:P2S());
        local precision = self.precision;
        --[[/* print zero bytes so we can specify NULL as target  */
        /* and get the length of the resulting str */]]
		local text = ("%.*f"):format(precision, number)
		local size = strlen(text)
        --/* number does not fit into field width: try to reduce precision */
        if (width < 0) then
            width = 0;
		end
        if (size > width and precision > 0) then
            local delta = size - width;
            if (delta > precision) then
                delta = precision;
			end
            precision = precision - delta;
            size = size - delta;
            --/* zero precision: omit decimal point, too */
            if (precision == 0) then
                size = size - 1
			end
        end
        ---/* number still doesn't fit: display '*****'  */
        if (size > width) then
            str.resize(width);
            for i = 0, width do
                str[i] = '*';
			end
        else
            str = text
        end
    end

    if str == "" or str ~= self.string then 
        update = update + 1;
        self.string = str;
    end

    --/* something has changed and should be updated */
    if (update > 0) then

		
        -- /* Init pingpong scroller. start scrolling left (wrong way) to get a delay */
        if (self.align == ALIGN_PINGPONG) then
            self.direction = 0;
            self.delay = PINGPONGWAIT;
        end
		--[[
        /* if there's a marquee scroller active, it has its own */
        /* update callback timer, so we do nothing here; otherwise */
        /* we simply call this scroll callback directly */
		]]
        if (self.align ~= ALIGN_MARQUEE or self.align ~= ALIGN_AUTOMATIC or self.align ~= ALIGN_PINGPONG) then
            self:TextScroll()
        end

    end

end
