
local MAJOR = "StarLibProperty-1.0" 
local MINOR = 1
assert(LibStub, MAJOR.." requires LibStub") 
local LibProperty = LibStub:NewLibrary(MAJOR, MINOR)
local libEval = "StarLibEvaluator-1.0"
assert(LibStub(libEval), MAJOR .. " requires " .. libEval)
local Evaluator = LibStub(libEval):New()
local errorHandler
assert(LibStub("StarLibError-1.0"), MAJOR .. " requires StarLibError-1.0")
local error = LibStub("StarLibError-1.0")
assert(error, "No error handler")
error = error:New(MAJOR)


if not LibProperty or not Evaluator then return end

if not LibProperty.pool then
	LibProperty.pool = setmetatable({},{__mode='k'})
	LibProperty.__index = LibProperty
end

function LibProperty:New(v, name, expression, defval)
	if not v or not expression then
		error:Print("Unable to create new property.")
		return
	end
	if not v or not name then return end
	
	local obj = next(self.pool)

	if obj then
		self.pool[obj] = nil
	else
		obj = {}
	end

	setmetatable(obj, self)

	v.visitor = v
	
	obj.expression = expression

	obj.is_valid = true
	
	if expression ~= nil and type(expression) == "string" then
		obj.result = Evaluator.ExecuteCode(v, name, expression, false, defval)
		if obj.result == nil then
			StarTip:Print(("Property: \"%s\""):format(expression))
			obj.is_valid = false
		end
	elseif expression ~= nil then
		StarTip:Print(("Property: <%s> has no expression."):format(name))
		obj.is_valid = false
	end
	
	obj.New = nil	
	
	return obj	
end

function LibProperty:Del(prop)
	if prop then
		LibProperty.pool[prop] = true
	else
		LibProperty.pool[self] = true
	end
end

function LibProperty:Eval()
	if not self.is_valid then return -1 end
	
	local update = 1
	
	local old = self.result
	
	self.result = Evaluator.ExecuteCode(self.visitor, self.expression)
	
	if old == self.result then
		update = 0
	end
	
	return update
end

function LibProperty:P2N()
	if not self.is_valid then return 0 end
	if type(self.result) == "number" then
		return self.result
	else
		return tonumber(self.result)
	end
end

function LibProperty:P2S()
	if not self.is_valid then return "" end
	
	if type(self.result) ~= "number" and type(self.result) ~= "string" then
		return ""
	end
	return ("%s"):format(self.result)
end
